// derivative.cpp

#include <stdio.h>
#include <math.h>
#if defined(WIN32) && !defined(NDEBUG)
#include <crtdbg.h>
#endif

#include "smile.h"


using namespace std;

void restoreColumn(DSL_Dmatrix &mtx, const vector<double> &origCol, int base)
{
	//printf("restore %d:", base);
	int outcomeCount = mtx.GetSizeOfDimension(mtx.GetLastDimension());
	for (int i = 0; i < outcomeCount; i ++)
	{
		mtx[base + i] = origCol[i];
		//printf(" %g", origCol[i]);
	}
	//printf("\n");
}

void changeProb(DSL_Dmatrix &mtx, int base, int outcome, double dp)
{
	int outcomeCount = mtx.GetSizeOfDimension(mtx.GetLastDimension());
	//printf("mod %d %g:", base, dp);

	//double total = 0;
	//for (int i = 0; i < outcomeCount; i ++) total += mtx[base + i];
	//double complement = total - mtx[base + outcome];

	double complement = 1.0 - mtx[base + outcome];
	
	for (int i = 0; i < outcomeCount; i ++)
	{
		double &p = mtx[base + i];
		if (i == outcome)
		{
			p += dp;
		}
		else
		{
			if (0 == complement)
			{
				p = -dp / (outcomeCount - 1);
			}
			else
			{
				p -= dp * p / complement;
			}
		}
		//printf(" %g", p);
	}
	//printf("\n");
}

double calcPosterior(DSL_network &net, int hTarget, int targetOutcome, int hSens, int cptIdx, double dp)
{
	DSL_Dmatrix &mtxSens = *(net.GetNode(hSens)->Definition()->GetMatrix());
	int outcomeCount = mtxSens.GetSizeOfDimension(mtxSens.GetLastDimension());
	int outcome = cptIdx % outcomeCount;
	int base = cptIdx - outcome;
	double *pData = mtxSens.GetItems().Items() + base;
	vector<double> origCol(pData, pData + outcomeCount);
	changeProb(mtxSens, base, outcome, dp);
	int res = net.UpdateBeliefs();
	assert(DSL_OKAY == res);
	restoreColumn(mtxSens, origCol, base);
	DSL_nodeValue *targetVal = net.GetNode(hTarget)->Value();
	double f = targetVal->GetMatrix()->Subscript(targetOutcome);
	return f;
}

void printNodeAndOutcome(DSL_network &net, int handle, int outcome)
{
	DSL_node *node = net.GetNode(handle);
	const char *nodeId = node->GetId();
	const char *outcomeId = node->Definition()->GetOutcomesNames()->Subscript(outcome);
	printf("%s=%s", nodeId, outcomeId);
}


void printCptAddress(DSL_network &net, int hSens, int index)
{
	DSL_Dmatrix &mtxSens = *(net.GetNode(hSens)->Definition()->GetMatrix());
	printf("%d: %g ", index, mtxSens[index]);
	
	DSL_intArray coords;
	mtxSens.IndexToCoordinates(index, coords);
	for (int i = 0; i < coords.GetSize(); i ++)
	{
		int c = coords[i];
		if (i == coords.GetSize() - 1)
		{
			printNodeAndOutcome(net, hSens, c);
		}
		else
		{
			printNodeAndOutcome(net, net.GetParents(hSens)[i], c);
		}
		printf("(%d) ", c);
	}
}

struct ErrorAccumulator
{
	ErrorAccumulator() : maxErr(0), maxErrSensNode(-1), maxErrSensCptIndex(-1), count(0), totalErr(0) {}
	void addError(double err, int hTargetNode, int targetOutcome, int hSensNode, int cptIdx)
	{
		totalErr += err;
		count ++;
		if (err > maxErr) 
		{
			maxErr = err;
			maxErrTargetNode = hTargetNode;
			maxErrTargetOutcome = targetOutcome;
			maxErrSensNode = hSensNode;
			maxErrSensCptIndex = cptIdx;
		}
	}

	void print(DSL_network &net, const char *prefix)
	{
		if (maxErr == 0)
		{
			printf("%s: NO ERROR\n\n", prefix);
		}
		else
		{
			printf("%s: avg:%g over %d\nmax:%.16g at [%d]%s ", prefix, totalErr / count, count, maxErr, maxErrSensNode, net.GetNode(maxErrSensNode)->GetId());
			printCptAddress(net, maxErrSensNode, maxErrSensCptIndex);
			DSL_node *pTarg = net.GetNode(maxErrTargetNode);
			printf("\nfor target [%d]%s : [%d]%s\n\n", maxErrTargetNode, pTarg->GetId(), maxErrTargetOutcome, 
				pTarg->Definition()->GetOutcomesNames()->Subscript(maxErrTargetOutcome));
		}
	}

	double maxErr;
	int maxErrTargetNode;
	int maxErrTargetOutcome;
	int maxErrSensNode;
	int maxErrSensCptIndex;
	double totalErr;
	int count;
};

void validateSensitivity(DSL_network &net, int targetNode, int targetOutcome, double dh, const DSL_sensitivity &sens, ErrorAccumulator &errGlobal, ErrorAccumulator &errGlobalDerivative, int &globalDeterministicCount)
{
	DSL_node *pTarg = net.GetNode(targetNode);
	printf("=== SENSITIVITY for target [%d]%s : [%d]%s\n\n", 
		targetNode, pTarg->GetId(), targetOutcome, 
		pTarg->Definition()->GetOutcomesNames()->Subscript(targetOutcome));

	ErrorAccumulator errPerTarget;
	ErrorAccumulator errPerTargetDerivative;

	for (int hSens = net.GetFirstNode(); hSens >= 0; hSens = net.GetNextNode(hSens))
	{
		DSL_node *node = net.GetNode(hSens);
		/*
		if (node->Value()->IsRealEvidence())
		{
			continue;
		}
		*/

		ErrorAccumulator nodeErr;
		
		DSL_Dmatrix &mtxSens = *(node->Definition()->GetMatrix());
		printf("--- %s [%d] mtx=%d\n", node->GetId(), node->Handle(), mtxSens.GetSize());
		
		int outcomeCount = mtxSens.GetSizeOfDimension(mtxSens.GetLastDimension());
		for (int base = 0; base < mtxSens.GetSize(); base += outcomeCount)
		{
			for (int outcome = 0; outcome < outcomeCount; outcome ++)
			{
				int cptIdx = base + outcome;
				printCptAddress(net, hSens, cptIdx);	
				printf("\n");
				double p0 = mtxSens[cptIdx];
				bool deterministic = false;

				double f_x_plus_h, f_x_minus_h, f_prim;
				/*
				double f_x_plus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, dh);
				double f_x_minus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, -dh);
				double f_x_plus_2h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, 2 * dh);
				double f_x_minus_2h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, -2 * dh);
				double f_prim = (-f_x_plus_2h + 8 * f_x_plus_h - 8 * f_x_minus_h + f_x_minus_2h) / (12 * dh);
				printf(" d4p: %.15g %.15g %.15g %.15g\n\t[ %.15g ]\n", f_x_minus_2h, f_x_minus_h, f_x_plus_h, f_x_plus_2h, f_prim);
				*/

				if (p0 == 0)
				{
					deterministic = true;
					f_x_plus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, 2 * dh);
					f_x_minus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, 0);
				}
				else if (p0 == 1)
				{
					deterministic = true;
					f_x_plus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, 0);
					f_x_minus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, -2 * dh);
				}
				else
				{
					f_x_plus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, dh);
					f_x_minus_h = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, -dh);
				}
				f_prim = (f_x_plus_h - f_x_minus_h) / (2 * dh);
				printf(" d2p: %.15g %.15g\n\t[ %.15g ]\n", f_x_minus_h, f_x_plus_h, f_prim);

				double f_prim_approx = f_prim;

				
				double p1, p2;
				if (fabs(0.5 - p0) > 0.25)
				{
					p1 = 0.375;
					p2 = 0.625;
				}
				else
				{
					p1 = 0.125;
					p2 = 0.875;
				}

				double q0 = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, 0);
				double q1 = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, p1 - p0);
				double q2 = calcPosterior(net, targetNode, targetOutcome, hSens, cptIdx, p2 - p0);

				double denom = -p0*q1+p0*q2+p1*q0-p1*q2-p2*q0+p2*q1;
				double a = (-p0*q0*q1+p0*q0*q2+p1*q0*q1-p1*q1*q2-p2*q0*q2+p2*q1*q2)/denom;
				double b = (-p0*p1*q0*q2+p0*p1*q1*q2+p0*p2*q0*q1-p0*p2*q1*q2-p1*p2*q0*q1+p1*p2*q0*q2)/denom;
				double c = (-p0*p1*q0+p0*p1*q1+p0*p2*q0-p0*p2*q2-p1*p2*q1+p1*p2*q2)/denom;

				if (fabs(denom) > 1e-10)
				{
					f_prim = (a * c - b) / ((p0 + c) * (p0 + c));
				}
				else
				{
					f_prim = (q1 - q0) / (p1 - p0);
				}

				printf(" eq3: %g:%g %g:%g %g:%g\n      denom=%g a=%g b=%g c=%g\n\t[ %.15g ]\n", p0, q0, p1, q1, p2, q2, denom, a, b, c, f_prim);

				DSL_sensitivity::Target sensTarget(targetNode, targetOutcome);
				const DSL_Dmatrix &sensDerivative = *sens.GetSensitivity(hSens, sensTarget);
				vector<const DSL_Dmatrix*> coeffs;
				sens.GetCoefficients(hSens, sensTarget, coeffs);
				
				double f_prim_sens, alpha, beta, gamma, delta;
				if (sensDerivative.GetSize() == 0)
				{
					f_prim_sens = alpha = beta = gamma = delta = 0;
				}
				else
				{
					f_prim_sens = sensDerivative[cptIdx];
					alpha = (*coeffs[0])[cptIdx];
					beta = (*coeffs[1])[cptIdx];
					gamma = (*coeffs[2])[cptIdx];
					delta = (*coeffs[3])[cptIdx];
				}
				
				printf(" sen: %g %g %g %g\n\t[ %.15g ]\n", alpha, beta, gamma, delta, f_prim_sens);

				if (deterministic)
				{
					globalDeterministicCount ++;
				}
				else
				{
					double err = fabs(f_prim - f_prim_sens);
					nodeErr.addError(err, targetNode, targetOutcome, hSens, cptIdx);
					errPerTarget.addError(err, targetNode, targetOutcome, hSens, cptIdx);
					errGlobal.addError(err, targetNode, targetOutcome, hSens, cptIdx);
				}
				double errDeriv = fabs(f_prim - f_prim_approx);
				errPerTargetDerivative.addError(errDeriv, targetNode, targetOutcome, hSens, cptIdx);
				errGlobalDerivative.addError(errDeriv, targetNode, targetOutcome, hSens, cptIdx);
				printf("\n");
			}
		}
		nodeErr.print(net, "NODE ERROR");
	}
	
	errPerTarget.print(net, "TARGET ERROR");
	errPerTargetDerivative.print(net, "TARGET ERROR FOR APPROX DIFFERENTIAL");
}



int doStuff(int argc, char *argv[])
{
	const double DIFF_H = 1e-5;

	DSL_network net;

	int res = net.ReadFile(argv[1]);
	if (DSL_OKAY != res)
	{
		return res;
	}
	
	vector<int> targetHandles;

	for (int i = 2; i < argc; i ++)
	{
		const char *pArg = argv[i];
		const char *pColon = strchr(pArg, ':');
		if (NULL == pColon)
		{
			int hTarget = net.FindNode(pArg);
			if (hTarget < 0)
			{
				return hTarget;
			}
			targetHandles.push_back(hTarget);
		}
		else
		{
			string id(pArg, pColon); 
			int hEvidence = net.FindNode(id.c_str());
			if (hEvidence < 0)
			{
				return hEvidence;
			}

			DSL_node *evNode = net.GetNode(hEvidence);
		
			pColon ++;
			int outcome;
			if (isdigit(*pColon))
			{
				outcome = atoi(pColon);
			}
			else
			{
				DSL_idArray *outIds = evNode->Definition()->GetOutcomesNames();
				outcome = outIds->FindPosition(pColon);
			}

			if (outcome < 0 || outcome >= evNode->Definition()->GetNumberOfOutcomes())
			{
				return DSL_WRONG_NUM_STATES;
			}

			res = evNode->Value()->SetEvidence(outcome);
			if (DSL_OKAY != res)
			{
				return res;
			}
		}
	}

	ErrorAccumulator errGlobal;
	ErrorAccumulator errGlobalDeriv;
	DSL_sensitivity sens;

	if (targetHandles.empty())
	{
		for (int h = net.GetFirstNode(); h >= 0; h = net.GetNextNode(h))
		{
			DSL_node *node = net.GetNode(h);
			if (!node->Value()->IsEvidence())
			{
				targetHandles.push_back(h);
			}
		}
	}
	
	net.ClearAllTargets();
	for (unsigned i = 0; i < targetHandles.size(); i ++)
	{
		net.SetTarget(targetHandles[i]);
	}
	res = sens.Calculate(net, true);
	assert(DSL_OKAY == res);
	
	int globalDeterministicCount = 0;
	for (unsigned i = 0; i < targetHandles.size(); i ++)
	{
		int h = targetHandles[i];
		DSL_node *node = net.GetNode(h);
		net.ClearAllTargets();
		net.SetTarget(h);
		//int res = sens.Calculate(net, true);
		//assert(DSL_OKAY == res);
		int outcomeCount = node->Definition()->GetNumberOfOutcomes();
		for (int outcome = 0; outcome < outcomeCount; outcome ++)
		{
			net.ClearAllTargets();
			net.SetTarget(h);
			validateSensitivity(net, h, outcome, DIFF_H, sens, errGlobal, errGlobalDeriv, globalDeterministicCount);
		}
	}

	printf("DETERMINISTIC CPT ENTRIES: %d\n\n", globalDeterministicCount);
	errGlobal.print(net, "GLOBAL ERROR");
	errGlobalDeriv.print(net, "GLOBAL ERROR FOR APPROX DIFFERENTIAL");

	printf("DIFF_H=%.16g\n", DIFF_H);

	return DSL_OKAY;
}


class MyErrRedir : public DSL_errorStringRedirect
{
	virtual void LogError(int code, const char *message) { printf("%d %s\n", code, message); }
};

int main(int argc, char *argv[])
{
#if defined(WIN32) && !defined(NDEBUG)
	_CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);
#endif
	MyErrRedir mer;
	ErrorH.Redirect(&mer);
	return doStuff(argc, argv);
}
